unit PanZoom;

{
 This unit contains the TPanZoom panel and its supporting classes.

 Copyright (c) 1997-2005 Dave Shapiro, Professional Software, Inc.
 Use and modify freely.
}

interface

(*$IFDEF Ver80*)
  (*$DEFINE PreDelphi3*)
  (*$DEFINE PreDelphi4*)
(*$ENDIF*)
(*$IFDEF Ver90*)
  (*$DEFINE PreDelphi3*)
  (*$DEFINE PreDelphi4*)
(*$ENDIF*)
(*$IFDEF Ver100*)
  (*$DEFINE PreDelphi4*)
(*$ENDIF*)

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, DoublePoints
  (*$IFDEF PreDelphi4*)
    , ProportionalScrollBar
  (*$ENDIF*)
  ;

const
  DefBandWidth = 1;
  crCrossPlus = 6;
  crCrossMinus = 7;
  crCrossZoom = 8;
  crCrossSelect = 9;
  crCrossErase = 10;
  crCrossLeftRight = 11;
  crEyeDropper = 12;
  crCrossBasic = 13;

type
  TRubberBandStopEvent = procedure(Sender: TObject; ARect: TRect) of object;

  TPsRubberBand = class(TComponent)
  private
  protected
    FCanvas: TCanvas;
    FProportional: Boolean;
    FPropWidth: Longint;
    FPropHeight: Longint;
    FOnStop: TRubberBandStopEvent;
    FOnCancel: TNotifyEvent;
    FHasStarted: Boolean;
    FBandWidth: Integer;
    Brush, OldBrush: TBrush;
    Band: TRect;
    procedure Paint;
    procedure UpdateBand(const X, Y: Integer);
    procedure SetBandWidth(const NewBandWidth: Integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CorrectBand: TRect;
    procedure Start(const X, Y: Integer);
    procedure Move(const X, Y: Integer);
    procedure Stop(const X, Y: Integer);
    procedure Cancel;
    procedure SetProportional(const APropWidth, APropHeight: Longint);
    procedure SetNonProportional;
    property Canvas: TCanvas read FCanvas write FCanvas;
    property Proportional: Boolean read FProportional;
    property HasStarted: Boolean read FHasStarted;
  published
    property OnStop: TRubberBandStopEvent read FOnStop write FOnStop;
    property OnCancel: TNotifyEvent read FOnCancel write FOnCancel;
    property BandWidth: Integer read FBandWidth write SetBandWidth
                                default DefBandWidth;
  end;

type
  TSpanningCursor = class(TObject)
  protected
    FHasPosition: Boolean;
    FPosition: TPoint;
    FPaintBox: TPaintBox;
    procedure Paint;
    procedure Clear;
  public
    constructor Create(PaintBox: TPaintBox);
    destructor Destroy; override;
    procedure Move(const Position: TPoint);
    procedure MoveSilent(const Position: TPoint);
  end;

type
  TScaleSystem = class(TPersistent)
  private
    { Given. }
    FMasterP1: TLongintPoint;    { Master value for first scale point. }
    FSkewP: TLongintPoint;       { Master value for skew point. }
    FMasterP2Prime: TLongintPoint; { Master value for second scale point. }
    FActualP1: TDoublePoint; { Real-world value of MasterP1, as given by user. }
    FActualP2: TDoublePoint; { Real-world value of MasterP2Prime, as given by user. }
    FIsLogarithmicP: TBooleanPoint;  { Whether each axis is logarithmic. }
    { Derived. }
    FCosTheta: Double;  { Cosine of skew angle. }
    FSinTheta: Double;  { Sine of skew angle. }
    FMasterP2: TLongintPoint; { Master value for second scale point, de-rotated by skew angle. }
    FDeltaMasterP: TDoublePoint;  { MasterP2 - MasterP1. }
    FUnitsPerPixelP: TDoublePoint; { Units per master pixel. }
    FXIsBackward: Boolean; { Whether or not the X axis goes more negative to the right. }
    FLogRatioP: TDoublePoint; { Pre-calculated for doing log axes quickly. }
  protected
  public
    WindowOriginP: TLongintPoint;       { Zooming stuff. }
    WindowExtentP: TLongintPoint;       {       |        }
    ViewportOriginP: TLongintPoint;     {       |        }
    ViewportExtentP: TLongintPoint;     {       |        }
    constructor Create;
    destructor Destroy; override;
    procedure AssignTo(Dest: TPersistent); override;
    procedure OneToOne;
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    procedure Define(const AMasterP1, ASkewP, AMasterP2Prime: TLongintPoint;
                     const AActualP1, AActualP2: TDoublePoint;
                     const AIsLogarithmicP: TBooleanPoint);
    function DeviceToMaster(const ADeviceP: TLongintPoint): TLongintPoint;
    function MasterToDevice(const AMasterP: TLongintPoint): TLongintPoint;
    function DeviceToActual(const ADeviceP: TLongintPoint): TDoublePoint;
    function ActualToDevice(const AActualP: TDoublePoint): TLongintPoint;
    function ActualToMaster(const AActualP: TDoublePoint): TLongintPoint;
    function MasterToActual(const MasterP: TLongintPoint): TDoublePoint;
    //function Transform(const P: TLongintPoint; const Dest: TScaleSystem): TLongintPoint;
    function PixelHeight: Integer;
    function PixelWidth: Integer;
    function HalfPixelHeight: Integer;
    function HalfPixelWidth: Integer;

    property MasterP1: TLongintPoint read FMasterP1;
    property SkewP: TLongintPoint read FSkewP;
    property MasterP2Prime: TLongintPoint read FMasterP2Prime;
    property ActualP1: TDoublePoint read FActualP1;
    property ActualP2: TDoublePoint read FActualP2;
    property IsLogarithmicP: TBooleanPoint read FIsLogarithmicP;
    property XIsBackward: Boolean read FXIsBackward;
  end;

const
  DefSizeHandleWidth = 6;
  DefMovable = True;
  DefSizable = True;

type
  TSizeHandleWidth = 0..MaxInt;

  TResizablePanel = class(TPanel)
  private
  protected
    FMovable: Boolean;
    FSizable: Boolean;
    FSizeHandleWidth: TSizeHandleWidth;
    procedure WMNCHitTest(var Mess: TWMNCHitTest); message WM_NCHITTEST;
    function GetTotalBorderWidth: Integer; virtual;
    function GetUsableWidth: Integer; virtual;
    function GetUsableHeight: Integer; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Movable: Boolean read FMovable write FMovable default DefMovable;
    property Sizable: Boolean read FSizable write FSizable default DefSizable;
    property SizeHandleWidth: TSizeHandleWidth read FSizeHandleWidth
                                               write FSizeHandleWidth
                                               default DefSizeHandleWidth;
    property TotalBorderWidth: Integer read GetTotalBorderWidth;
    property UsableWidth: Integer read GetUsableWidth;
    property UsableHeight: Integer read GetUsableHeight;
  end;

type
  TMouseEnterPaintBox = class(TPaintBox)
  protected
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    procedure CMMouseEnter(var Msg: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Msg: TMessage); message CM_MOUSELEAVE;
  published
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter
                                        write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave
                                        write FOnMouseLeave;
  end;

const
  DefSmallChange = 2;
  DefLargeChange = 10;
  DefPaintBoxCursor = crCrossBasic;
  DefZoomCursor = crCrossZoom;
  DefUseFastPainting = False;
  DefUseSpanningCursor = False;
  
type
  TPageSizeScrollBar =
  (*$IFDEF PreDelphi4*)
    TPsProportionalScrollBar
  (*$ELSE*)
    TScrollBar
  (*$ENDIF*)
  ;

  TScalePoint = record
    MasterP: TLongintPoint;
    ActualP: TDoublePoint;
  end;

  TDefinableScalePoint = record
    MasterP: TLongintPoint;
    ActualP: TDoublePoint;
    Defined: Boolean;
  end;

  TPanZoomMouseMoveEvent = procedure(Sender: TObject; Shift: TShiftState;
                                     const DeviceP, MasterP: TLongintPoint) of object;

  TPanZoomMouseEvent = procedure(Sender: TObject; Button: TMouseButton;
                          Shift: TShiftState; DeviceX, DeviceY: Integer;
                          ActualX, ActualY: Double) of object;

  TPanZoomSelectEvent = procedure(Sender: TObject; DeviceSelection: TRect;
                                  ActualSelection: TQuadrilateral) of object;

  EDefineScale = class(Exception);

  TPanZoomPanel = class;

  TPanZoomPanelStateItem = (siMovable, siPainting, siPanning, siZooming, siSelecting, siPaintBoxCursor);

  TPanZoomPanelState = class(TObject)
  protected
    FMovable: Boolean;
    FIsPainting: Boolean;
    FIsPenDown: Boolean;
    FIsPanning: Boolean;
    FIsZooming: Boolean;
    FIsSelecting: Boolean;
    FPaintBoxCursor: TCursor;
    FUseSpanningCursor: Boolean;
  public
    procedure Read(Panel: TPanZoomPanel);
  end;

  IPanZoomPanelObserver = interface(IUnknown) ['{7CFA3F7E-5387-410D-BF95-EABA37DDB8EC}']
    procedure OnPaintBoxPaint(Panel: TPanZoomPanel);
    procedure OnPenDownChange(Panel: TPanZoomPanel);
    procedure OnPenWidthChange(Panel: TPanZoomPanel);
  end;

  { HERE IT IS! }

  TPanZoomPanel = class(TResizablePanel)
  protected
    FObservers: TInterfaceList;
    FSelectCompleted: Boolean;
    FStates: TList;
    PrevX: Integer;
    PrevY: Integer;
    FOriginalScaleSystem, FCurrentScaleSystem: TScaleSystem;
    FirstPoint: Boolean;
    FScaleDefined: Boolean;
    FBitmap: TBitmap;
    FPaintBox: TMouseEnterPaintBox;
    FRubberBand: TPsRubberBand;
    FHorzScrollBar, FVertScrollBar: TPageSizeScrollBar;
    FIsZoomingIn: Boolean;
    FUseProportionalRubberBand: Boolean;
    FZoomCursor: TCursor;
    FIsPainting: Boolean;
    FIsPanning: Boolean;
    FIsSelecting: Boolean;
    FPenDown: Boolean;
    FUseFastPainting: Boolean;
    FFirstScalePoint: TDefinableScalePoint;
    FSkewPoint: TDefinableScalePoint;
    FSecondScalePoint: TDefinableScalePoint;
    FShowScalePoints: Boolean;
    FSpanningCursor: TSpanningCursor;
    FUseSpanningCursor: Boolean;
    FOnResize: TNotifyEvent;
    FOnPaintBoxMouseDown: TPanZoomMouseEvent;
    FOnPaintBoxMouseMove: TPanZoomMouseMoveEvent;
    FOnPaintBoxMouseUp: TPanZoomMouseEvent;
    FOnPaintBoxMouseEnter: TNotifyEvent;
    FOnPaintBoxMouseLeave: TNotifyEvent;
    FOnPaintBoxClick: TNotifyEvent;
    FOnPaintBoxPaint: TNotifyEvent;
    FOnZoomInBegin: TNotifyEvent;
    FOnZoomInEnd: TNotifyEvent;
    FOnZoomToArea: TNotifyEvent;
    FOnZoomInCancel: TNotifyEvent;
    FOnSelectBegin: TNotifyEvent;
    FOnSelectEnd: TNotifyEvent;
    FOnSelectArea: TPanZoomSelectEvent;
    FOnSelectCancel: TNotifyEvent;
    FOnZoomOut: TNotifyEvent;
    FOnPaintBegin: TNotifyEvent;
    FOnPaintEnd: TNotifyEvent;
    FOnPanBegin: TNotifyEvent;
    FOnPanEnd: TNotifyEvent;
    FOnPenChanged: TNotifyEvent;
    FOnScaleDefined: TNotifyEvent;
    FOnScaleUndefined: TNotifyEvent;
    procedure CheckZoomFactor;
    procedure ScrollHorzOrVert(const Delta: Integer; ScrollBar: TPageSizeScrollBar);
    procedure ClearPrevCoords;
    procedure ReconfigureCurrentSystem;
    procedure SetupScrollBars;
    procedure PaintBitmap;
    procedure PaintBoxPaint(Sender: TObject); virtual;
    procedure PaintBoxMouseMove(Sender: TObject; Shift: TShiftState;
                                X, Y: Integer);
    procedure RubberBandStop(Sender: TObject; ARect: TRect);
    procedure ZoomToArea(ARect: TRect);
    procedure HorzScrollBarScroll(Sender: TObject; ScrollCode: TScrollCode;
                                  var ScrollPos: Integer);
    procedure VertScrollBarScroll(Sender: TObject; ScrollCode: TScrollCode;
                                  var ScrollPos: Integer);
    procedure DoResize;
    procedure SetBitmap(const NewBitmap: TBitmap); virtual;
    procedure Paint; override;
    procedure PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
                                Shift: TShiftState; X, Y: Integer); virtual;
    procedure PaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
                              Shift: TShiftState; X, Y: Integer); virtual;
    procedure PaintBoxMouseEnter(Sender: TObject);
    procedure PaintBoxMouseLeave(Sender: TObject);
    procedure PaintBoxClick(Sender: TObject);
    procedure StoreState; virtual;
    procedure RestoreState(const IgnoreItem: TPanZoomPanelStateItem); virtual;
    procedure InternalSetIsZoomingIn(const Value: Boolean);
    procedure InternalSetIsPainting(const Value: Boolean);
    procedure InternalSetIsPanning(const Value: Boolean);
    procedure SetIsZoomingIn(const Value: Boolean);
    procedure SetIsPainting(const Value: Boolean);
    procedure SetIsPanning(const Value: Boolean);
    procedure SetIsSelecting(const Value: Boolean);
    procedure InternalSetIsSelecting(const Value: Boolean);
    procedure SetUseProportionalRubberBand(const Value: Boolean);
    procedure SetPaintBoxTransformation;
    procedure UnsetPaintBoxTransformation;
    procedure SetPenDown(const Value: Boolean);
    procedure CreateAndSetPaintingCursor;
    function GetPenWidth: Integer;
    procedure SetPenWidth(Value: Integer);
    procedure CheckPenWidth;
    function GetRubberBandWidth: Integer;
    procedure SetRubberBandWidth(const NewRubberBandWidth: Integer);
    procedure Resize; override;
    function GetPaintBoxCursor: TCursor;
    procedure RedrawCursor;
    procedure SetPaintBoxCursor(const Value: TCursor);
    function GetPaintBoxCanvas: TCanvas;
    procedure SetUseFastPainting(const Value: Boolean);
    procedure ConvertIfFastPainting;
    procedure Loaded; override;
    procedure SetSmallChange(const Value: Integer);
    procedure SetLargeChange(const Value: Integer);
    function GetSmallChange: Integer;
    function GetLargeChange: Integer;
    procedure SetShowScalePoints(const Value: Boolean);
    procedure PaintScalePoints;
    procedure DoPaintScalePoints;
    procedure SetUseSpanningCursor(const Value: Boolean);
    procedure DoEnter; override;
    function IsObserved(const Observer: IPanZoomPanelObserver): Boolean;
    procedure BroadcastPaintBoxPaint;
    procedure BroadcastPenDownChange;
    procedure BroadcastPenWidthChange;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Attach(const Observer: IPanZoomPanelObserver);
    procedure Detach(const Observer: IPanZoomPanelObserver);
    procedure UpdatePaintingCursor;
    procedure ScrollHorz(const Delta: Longint);
    procedure ScrollVert(const Delta: Longint);
    procedure ClearScalePoints;
    procedure ZoomOut(const Factor: Double);
    procedure OneToOne;
    procedure SizeToImage;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure DefineScale(const XIsLogarithmic, YIsLogarithmic: Boolean);
    procedure SetScaleSystems(const Original, Current: TScaleSystem);
    procedure DefineScaleNoSkew(const XIsLogarithmic, YIsLogarithmic: Boolean);
    procedure UndefineScale;
    procedure ClearBitmap;
    procedure SetBitmapRef(const NewBitmap: TBitmap);
    procedure SetFirstScalePoint(const NewFirstDevice: TLongintPoint;
                                 const NewFirstActual: TDoublePoint);
    procedure UnsetFirstScalePoint;
    procedure SetSkewPoint(const NewSkewPoint: TLongintPoint);
    procedure UnsetSkewPoint;
    procedure SetSecondScalePoint(const NewSecondDevice: TLongintPoint;
                                  const NewSecondActual: TDoublePoint);
    procedure UnsetSecondScalePoint;

    function ActualToDevice(const ActualP: TDoublePoint): TLongintPoint;
    function DeviceToActual(const DeviceP: TLongintPoint): TDoublePoint;
    //function MaxDeviceP: TLongintPoint;
    function IsRubberBanding: Boolean;

    property Bitmap: TBitmap read FBitmap write SetBitmap;
    property PaintBoxCanvas: TCanvas read GetPaintBoxCanvas;
    property IsPainting: Boolean read FIsPainting write SetIsPainting;
    property IsPanning: Boolean read FIsPanning write SetIsPanning;
    property PenWidth: Integer read GetPenWidth write SetPenWidth;
    property FirstScalePointMaster: TLongintPoint read FFirstScalePoint.MasterP;
    property FirstScalePointActual: TDoublePoint read FFirstScalePoint.ActualP;
    property FirstScalePointDefined: Boolean read FFirstScalePoint.Defined;
    property PenDown: Boolean read FPenDown write SetPenDown;
    property ScaleDefined: Boolean read FScaleDefined;
    property SecondScalePointActual: TDoublePoint read FSecondScalePoint.ActualP;
    property SecondScalePointDefined: Boolean read FSecondScalePoint.Defined;
    property SecondScalePointMaster: TLongintPoint read FSecondScalePoint.MasterP;
    property SkewPointDefined: Boolean read FSkewPoint.Defined;
    property SkewPointMaster: TLongintPoint read FSkewPoint.MasterP;
    property IsZoomingIn: Boolean read FIsZoomingIn write SetIsZoomingIn;
    property IsSelecting: Boolean read FIsSelecting write SetIsSelecting;
    {property IsLogarthmic: TBooleanPoint read FIsLogarithmic;}
    property OriginalScaleSystem: TScaleSystem read FOriginalScaleSystem;
    property CurrentScaleSystem: TScaleSystem read FCurrentScaleSystem;
    property PaintBox: TMouseEnterPaintBox read FPaintBox;
  published
    {property Bitmap: TBitmap read FBitmap write SetBitmap;}
    property LargeChange: Integer read GetLargeChange write SetLargeChange
                                  default DefLargeChange;
    property PaintBoxCursor: TCursor read GetPaintBoxCursor
                                     write SetPaintBoxCursor
                                     default DefPaintBoxCursor;
    property UseProportionalRubberBand: Boolean read FUseProportionalRubberBand
                                                write SetUseProportionalRubberBand;
    property RubberBandWidth: Integer read GetRubberBandWidth
                                      write SetRubberBandWidth;
    property ShowScalePoints: Boolean read FShowScalePoints
                                      write SetShowScalePoints;
    property SmallChange: Integer read GetSmallChange write SetSmallChange
                                  default DefSmallChange;
    property UseFastPainting: Boolean read FUseFastPainting
                                      write SetUseFastPainting
                                      default DefUseFastPainting;
    property UseSpanningCursor: Boolean read FUseSpanningCursor
                                        write SetUseSpanningCursor
                                        default DefUseSpanningCursor;
    property ZoomCursor: TCursor read FZoomCursor write FZoomCursor
                                 default DefZoomCursor;

    { Events. }
    property OnClick: TNotifyEvent read FOnPaintBoxClick write FOnPaintBoxClick;
    property OnMouseDown: TPanZoomMouseEvent read FOnPaintBoxMouseDown
                                             write FOnPaintBoxMouseDown;
    property OnMouseEnter: TNotifyEvent read FOnPaintBoxMouseEnter
                                        write FOnPaintBoxMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnPaintBoxMouseLeave
                                        write FOnPaintBoxMouseLeave;
    property OnMouseMove: TPanZoomMouseMoveEvent read FOnPaintBoxMouseMove
                                                 write FOnPaintBoxMouseMove;
    property OnMouseUp: TPanZoomMouseEvent read FOnPaintBoxMouseUp
                                           write FOnPaintBoxMouseUp;
    property OnPaintBegin: TNotifyEvent read FOnPaintBegin write FOnPaintBegin;
    property OnPaintEnd: TNotifyEvent read FOnPaintEnd write FOnPaintEnd;
    property OnPanBegin: TNotifyEvent read FOnPanBegin write FOnPanBegin;
    property OnPanEnd: TNotifyEvent read FOnPanEnd write FOnPanEnd;
    property OnPaintBoxPaint: TNotifyEvent read FOnPaintBoxPaint write FOnPaintBoxPaint;
    property OnPenChanged: TNotifyEvent read FOnPenChanged write FOnPenChanged;    
    property OnScaleDefined: TNotifyEvent read FOnScaleDefined write FOnScaleDefined;
    property OnScaleUndefined: TNotifyEvent read FOnScaleUndefined write FOnScaleUndefined;
    property OnZoomInBegin: TNotifyEvent read FOnZoomInBegin write FOnZoomInBegin;
    property OnZoomInCancel: TNotifyEvent read FOnZoomInCancel write FOnZoomInCancel;
    property OnZoomInEnd: TNotifyEvent read FOnZoomInEnd write FOnZoomInEnd;
    property OnZoomToArea: TNotifyEvent read FOnZoomToArea write FOnZoomToArea;
    property OnZoomOut: TNotifyEvent read FOnZoomOut write FOnZoomOut;
    property OnSelectBegin: TNotifyEvent read FOnSelectBegin write FOnSelectBegin;
    property OnSelectEnd: TNotifyEvent read FOnSelectEnd write FOnSelectEnd;
    property OnSelectArea: TPanZoomSelectEvent read FOnSelectArea write FOnSelectArea;
    property OnSelectCancel: TNotifyEvent read FOnSelectCancel write FOnSelectCancel;
  end;

procedure Register;

implementation

{$R PanZoomCursors.res}

uses Math;

const
  crPainting = 5;
  UndefinedLongintPoint: TLongintPoint = (X: 0; Y: 0);
  UndefinedDoublePoint: TDoublePoint = (X: 0.0; Y: 0.0);

var
  (*$IFDEF Win32*)
    OsVersionInfo: TOsVersionInfo;
  (*$ENDIF*)
  RunningNT: Boolean;
  (*$IFDEF PreDelphi3*)
    StupidHandle: HDC;
  (*$ENDIF*)

procedure Register;
begin
  RegisterComponents('Professional Software',
                     [TPsRubberBand, TResizablePanel, TPanZoomPanel]);
end;

procedure SetPen(const ACanvas: TCanvas; const AColor: TColor;
                 const AMode: TPenMode; const AStyle: TPenStyle;
                 const AWidth: Integer);
begin
  with ACanvas.Pen do begin
    Color := AColor;
    Mode := AMode;
    Style := AStyle;
    Width := AWidth;
  end;
end;

procedure SetBrush(const ACanvas: TCanvas; const AColor: TColor;
                   const AStyle: TBrushStyle);
begin
  with ACanvas.Brush do begin
    Color := AColor;
    Style := AStyle;
  end;
end;

procedure SetPenAndBrush(const ACanvas: TCanvas; const APenColor: TColor;
                         const AMode: TPenMode; const APenStyle: TPenStyle;
                         const AWidth: Integer; const ABrushColor: TColor;
                         const ABrushStyle: TBrushStyle);
begin
  SetPen(ACanvas, APenColor, AMode, APenStyle, AWidth);
  SetBrush(ACanvas, ABrushColor, ABrushStyle);
end;


{ -----------------------------------------------------------------------------}


constructor TScaleSystem.Create;
begin
  inherited;
  OneToOne;
end;

destructor TScaleSystem.Destroy;
begin
  inherited;
end;

procedure TScaleSystem.AssignTo(Dest: TPersistent);
begin
  if Dest is TScaleSystem then with TScaleSystem(Dest) do begin
    FMasterP1 := Self.FMasterP1;
    FSkewP := Self.FSkewP;
    FMasterP2Prime := Self.FMasterP2Prime;
    FActualP1 := Self.FActualP1;
    FActualP2 := Self.FActualP2;
    FIsLogarithmicP := Self.FIsLogarithmicP;
    FCosTheta := Self.FCosTheta;
    FSinTheta := Self.FSinTheta;
    FMasterP2 := Self.FMasterP2;
    FDeltaMasterP := Self.FDeltaMasterP;
    FUnitsPerPixelP := Self.FUnitsPerPixelP;
    FXIsBackward := Self.FXIsBackward;
    FLogRatioP := Self.FLogRatioP;
    WindowOriginP := Self.WindowOriginP;
    WindowExtentP := Self.WindowExtentP;
    ViewportOriginP := Self.ViewportOriginP;
    ViewportExtentP := Self.ViewportExtentP;
  end
  else begin
    inherited;
  end;
end;

procedure TScaleSystem.OneToOne;
begin
  WindowOriginP.X := 0;
  WindowOriginP.Y := 0;
  WindowExtentP.X := 1;
  WindowExtentP.Y := 1;
  ViewportOriginP.X := 0;
  ViewportOriginP.Y := 0;
  ViewportExtentP.X := 1;
  ViewportExtentP.Y := 1;
end;

type
  EScaleSystem = class(Exception);

procedure TScaleSystem.LoadFromStream(Stream: TStream);
begin
  Stream.Read(FMasterP1, SizeOf(FMasterP1));
  Stream.Read(FSkewP, SizeOf(FSkewP));
  Stream.Read(FMasterP2Prime, SizeOf(FMasterP2Prime));
  Stream.Read(FActualP1, SizeOf(FActualP1));
  Stream.Read(FActualP2, SizeOf(FActualP2));
  Stream.Read(FIsLogarithmicP, SizeOf(FIsLogarithmicP));
  Stream.Read(WindowOriginP, SizeOf(WindowOriginP));
  Stream.Read(WindowExtentP, SizeOf(WindowExtentP));
  Stream.Read(ViewportOriginP, SizeOf(ViewportOriginP));
  Stream.Read(ViewportExtentP, SizeOf(ViewportExtentP));
  Define(FMasterP1, FSkewP, FMasterP2Prime, FActualP1, FActualP2, FIsLogarithmicP);
end;

procedure TScaleSystem.SaveToStream(Stream: TStream);
begin
  Stream.Write(FMasterP1, SizeOf(FMasterP1));
  Stream.Write(FSkewP, SizeOf(FSkewP));
  Stream.Write(FMasterP2Prime, SizeOf(FMasterP2Prime));
  Stream.Write(FActualP1, SizeOf(FActualP1));
  Stream.Write(FActualP2, SizeOf(FActualP2));
  Stream.Write(FIsLogarithmicP, SizeOf(FIsLogarithmicP));
  Stream.Write(WindowOriginP, SizeOf(WindowOriginP));
  Stream.Write(WindowExtentP, SizeOf(WindowExtentP));
  Stream.Write(ViewportOriginP, SizeOf(ViewportOriginP));
  Stream.Write(ViewportExtentP, SizeOf(ViewportExtentP));
end;

procedure TScaleSystem.Define(const AMasterP1, ASkewP, AMasterP2Prime: TLongintPoint;
                              const AActualP1, AActualP2: TDoublePoint;
                              const AIsLogarithmicP: TBooleanPoint);
const
  TOLERANCE = 1e-30;
var
  Theta: Double;
begin
  if (AIsLogarithmicP.X and ((AActualP1.X <= 0) or (AActualP2.X <= 0))) or
     (AIsLogarithmicP.Y and ((AActualP1.Y <= 0) or (AActualP2.Y <= 0))) then
  begin
    raise EScaleSystem.Create('Logarithmic scale values must be positive.');
  end;
  if (Abs(AActualP1.X - AActualP2.X) < TOLERANCE) or
     (Abs(AActualP1.Y - AActualP2.Y) < TOLERANCE) then
  begin
    raise EScaleSystem.Create('Scale values must be different between points.');
  end;
  FMasterP1 := AMasterP1;
  FSkewP := ASkewP;
  FMasterP2Prime := AMasterP2Prime;
  FActualP1 := AActualP1;
  FActualP2 := AActualP2;
  FIsLogarithmicP := AIsLogarithmicP;
  Theta := ArcTan((FSkewP.Y - FMasterP1.Y) / (FSkewP.X - FMasterP1.X));
  FCosTheta := Cos(Theta);
  FSinTheta := Sin(Theta);
  FMasterP2.X := Round( FMasterP1.X + (FMasterP2Prime.X - FMasterP1.X) *
                 FCosTheta + (FMasterP2Prime.Y - FMasterP1.Y) * FSinTheta );
  FMasterP2.Y := Round( FMasterP1.Y + (FMasterP2Prime.Y - FMasterP1.Y) *
                 FCosTheta - (FMasterP2Prime.X - FMasterP1.X) * FSinTheta );
  FDeltaMasterP.X := FMasterP2.X - FMasterP1.X;
  FDeltaMasterP.Y := FMasterP2.Y - FMasterP1.Y;
  FUnitsPerPixelP.X := (FActualP2.X - FActualP1.X) / FDeltaMasterP.X;
  FUnitsPerPixelP.Y := (FActualP2.Y - FActualP1.Y) / FDeltaMasterP.Y;
  FXIsBackward := FUnitsPerPixelP.X < 0;
  if FIsLogarithmicP.X then FLogRatioP.X := Ln(FActualP2.X / FActualP1.X);
  if FIsLogarithmicP.Y then FLogRatioP.Y := Ln(FActualP2.Y / FActualP1.Y);
end;

function TScaleSystem.DeviceToMaster(const ADeviceP: TLongintPoint): TLongintPoint;
  // Requires only Window origin & extent, and Viewport origin & extent.
begin
  Result.X := WindowOriginP.X +
              MulDiv(ADeviceP.X - HalfPixelWidth - ViewportOriginP.X,
                     WindowExtentP.X, ViewportExtentP.X);
  Result.Y := WindowOriginP.Y +
              MulDiv(ADeviceP.Y - HalfPixelHeight - ViewportOriginP.Y,
                     WindowExtentP.Y, ViewportExtentP.Y);
end;

function TScaleSystem.MasterToDevice(const AMasterP: TLongintPoint): TLongintPoint;
  // Requires only Window origin & extent, and Viewport origin & extent.
begin
  Result.X := ViewportOriginP.X + HalfPixelWidth +
              MulDiv(AMasterP.X - WindowOriginP.X, ViewportExtentP.X, WindowExtentP.X);
  Result.Y := ViewportOriginP.Y + HalfPixelHeight +
              MulDiv(AMasterP.Y - WindowOriginP.Y, ViewportExtentP.Y, WindowExtentP.Y);
end;

function TScaleSystem.DeviceToActual(const ADeviceP: TLongintPoint): TDoublePoint;
var
  MasterP: TLongintPoint;
begin
  // Device to master.
  MasterP := DeviceToMaster(ADeviceP);
  // Master to actual.
  Result := MasterToActual(MasterP);
end;

function TScaleSystem.ActualToDevice(const AActualP: TDoublePoint): TLongintPoint;
var
  MasterP: TLongintPoint;
begin
  // Actual to master.
  MasterP := ActualToMaster(AActualP);
  // Master to device.
  Result := MasterToDevice(MasterP);
end;

function TScaleSystem.ActualToMaster(const AActualP: TDoublePoint): TLongintPoint;
var
  PPrimePixels: TDoublePoint;
begin
  // Actual to master.
  if FIsLogarithmicP.X then
    PPrimePixels.X := Ln(AActualP.X / FActualP1.X) / FLogRatioP.X * FDeltaMasterP.X
  else
    PPrimePixels.X := (AActualP.X - FActualP1.X) / FUnitsPerPixelP.X;
  if FIsLogarithmicP.Y then
    PPrimePixels.Y := Ln(AActualP.Y / FActualP1.Y) / FLogRatioP.Y * FDeltaMasterP.Y
  else
    PPrimePixels.Y := (AActualP.Y - FActualP1.Y) / FUnitsPerPixelP.Y;
  Result.X := Round(PPrimePixels.X * FCosTheta - PPrimePixels.Y * FSinTheta) + FMasterP1.X;
  Result.Y := Round(PPrimePixels.Y * FCosTheta + PPrimePixels.X * FSinTheta) + FMasterP1.Y;
end;

function TScaleSystem.MasterToActual(const MasterP: TLongintPoint): TDoublePoint;
var
  PDiffPixels, PPrimePixels: TDoublePoint;
begin
  // Master to actual.
  PDiffPixels.X := MasterP.X - FMasterP1.X;
  PDiffPixels.Y := MasterP.Y - FMasterP1.Y;
  PPrimePixels.X := PDiffPixels.X * FCosTheta + PDiffPixels.Y * FSinTheta;
  PPrimePixels.Y := PDiffPixels.Y * FCosTheta - PDiffPixels.X * FSinTheta;
  if FIsLogarithmicP.X then
    Result.X := FActualP1.X * Exp(PPrimePixels.X / FDeltaMasterP.X * FLogRatioP.X)
  else
    Result.X := FActualP1.X + FUnitsPerPixelP.X * PPrimePixels.X;
  if FIsLogarithmicP.Y then
    Result.Y := FActualP1.Y * Exp(PPrimePixels.Y / FDeltaMasterP.Y * FLogRatioP.Y)
  else
    Result.Y := FActualP1.Y + FUnitsPerPixelP.Y * PPrimePixels.Y;
end;

{
function TScaleSystem.Transform(const P: TLongintPoint;
                                const Dest: TScaleSystem): TLongintPoint;
begin
  Result.X := MulDiv(P.X,
                     MulDiv(Self.WindowExtentP.X, Dest.ViewportExtentP.X, Self.ViewportExtentP.X),
                     Dest.WindowExtentP.X);
  Result.Y := MulDiv(P.Y,
                     MulDiv(Self.WindowExtentP.Y, Dest.ViewportExtentP.Y, Self.ViewportExtentP.Y),
                     Dest.WindowExtentP.Y);
end;
}

function TScaleSystem.PixelHeight: Integer;
begin
  Result := Round(ViewportExtentP.Y / WindowExtentP.Y);
end;

function TScaleSystem.PixelWidth: Integer;
begin
  Result := Round(ViewportExtentP.X / WindowExtentP.X);
end;

function TScaleSystem.HalfPixelHeight: Integer;
begin
  Result := Round(ViewportExtentP.Y / (WindowExtentP.Y * 2));
end;

function TScaleSystem.HalfPixelWidth: Integer;
begin
  Result := Round(ViewportExtentP.X / (WindowExtentP.X * 2));
end;

{ -----------------------------------------------------------------------------}


procedure TMouseEnterPaintBox.CMMouseEnter(var Msg: TMessage);
begin
  inherited;
  if Assigned(FOnMouseEnter) then FOnMouseEnter(Self);
end;

procedure TMouseEnterPaintBox.CMMouseLeave(var Msg: TMessage);
begin
  inherited;
  if Assigned(FOnMouseLeave) then FOnMouseLeave(Self);
end;


{ -----------------------------------------------------------------------------}

procedure LoadCursors;
const
  CursorsLoaded: Boolean = False;
var
  CursorHandle: HCURSOR;
begin
  if not CursorsLoaded then begin
    CursorHandle := LoadCursor(HInstance, 'CROSSPLUS');
    if CursorHandle <> 0 then Screen.Cursors[crCrossPlus] := CursorHandle;
    CursorHandle := LoadCursor(HInstance, 'CROSSMINUS');
    if CursorHandle <> 0 then Screen.Cursors[crCrossMinus] := CursorHandle;
    CursorHandle := LoadCursor(HInstance, 'CROSSZOOM');
    if CursorHandle <> 0 then Screen.Cursors[crCrossZoom] := CursorHandle;
    CursorHandle := LoadCursor(HInstance, 'CROSSSELECT');
    if CursorHandle <> 0 then Screen.Cursors[crCrossSelect] := CursorHandle;
    CursorHandle := LoadCursor(HInstance, 'CROSSERASE');
    if CursorHandle <> 0 then Screen.Cursors[crCrossErase] := CursorHandle;
    CursorHandle := LoadCursor(HInstance, 'CROSSLEFTRIGHT');
    if CursorHandle <> 0 then Screen.Cursors[crCrossLeftRight] := CursorHandle;
    CursorHandle := LoadCursor(HInstance, 'EYEDROPPER');
    if CursorHandle <> 0 then Screen.Cursors[crEyeDropper] := CursorHandle;
    CursorHandle := LoadCursor(HInstance, 'CROSSBASIC');
    if CursorHandle <> 0 then Screen.Cursors[crCrossBasic] := CursorHandle;
    CursorsLoaded := True;
  end;
end;

constructor TPanZoomPanel.Create(AOwner: TComponent);
begin
  inherited;
  if not (csDesigning in ComponentState) then LoadCursors;
  FObservers := TInterfaceList.Create;
  FStates := TList.Create;
  FShowScalePoints := False;
  FCurrentScaleSystem := TScaleSystem.Create;
  FOriginalScaleSystem := TScaleSystem.Create;
  ClearScalePoints;
  FZoomCursor := DefZoomCursor;
  FUseFastPainting := DefUseFastPainting;
  FUseSpanningCursor := DefUseSpanningCursor;
  FHorzScrollBar := TPageSizeScrollBar.Create(Self);
  FHorzScrollBar.Parent := Self;
  FHorzScrollBar.Kind := sbHorizontal;
  FHorzScrollBar.OnScroll := HorzScrollBarScroll;
  FHorzScrollBar.TabStop := False;
  FVertScrollBar := TPageSizeScrollBar.Create(Self);
  FVertScrollBar.Parent := Self;
  FVertScrollBar.Kind := sbVertical;
  FVertScrollBar.OnScroll := VertScrollBarScroll;
  FVertScrollBar.TabStop := False;
  SetSmallChange(DefSmallChange);
  SetLargeChange(DefLargeChange);
  FPaintBox := TMouseEnterPaintBox.Create(Self);
  FPaintBox.Parent := Self;
  FPaintBox.ControlStyle := FPaintBox.ControlStyle + [csOpaque];
  FPaintBox.Align := alClient;
  FPaintBox.OnPaint := PaintBoxPaint;
  FPaintBox.OnMouseMove := PaintBoxMouseMove;
  FPaintBox.OnMouseDown := PaintBoxMouseDown;
  FPaintBox.OnMouseUp := PaintBoxMouseUp;
  FPaintBox.OnMouseEnter := PaintBoxMouseEnter;
  FPaintBox.OnMouseLeave := PaintBoxMouseLeave;
  FPaintBox.OnClick := PaintBoxClick;
  SetPaintBoxCursor(DefPaintBoxCursor);
  ClearPrevCoords;
  FRubberBand := TPsRubberBand.Create(Self);
  FRubberBand.Canvas := FPaintBox.Canvas;
  FRubberBand.OnStop := Self.RubberBandStop;
end;

destructor TPanZoomPanel.Destroy;
var
  I: Integer;
begin
  SetUseSpanningCursor(False);
  ClearBitmap;
  FRubberBand.Free;
  FPaintBox.Free;
  FVertScrollBar.Free;
  FHorzScrollBar.Free;
  OriginalScaleSystem.Free;
  CurrentScaleSystem.Free;
  for I := 0 to FStates.Count - 1 do TObject(FStates[I]).Free;
  FStates.Free;
  FObservers.Free;
  inherited;
end;

procedure TPanZoomPanel.Loaded;
begin
  inherited;
end;

procedure TPanZoomPanel.SetSmallChange(const Value: Integer);
begin
  FHorzScrollBar.SmallChange := Value;
  FVertScrollBar.SmallChange := Value;
end;

procedure TPanZoomPanel.SetLargeChange(const Value: Integer);
begin
  FHorzScrollBar.LargeChange := Value;
  FVertScrollBar.LargeChange := Value;
end;

function TPanZoomPanel.GetSmallChange: Integer;
begin
  Result := FHorzScrollBar.SmallChange;
end;

function TPanZoomPanel.GetLargeChange: Integer;
begin
  Result := FHorzScrollBar.LargeChange;
end;

procedure TPanZoomPanel.SetFirstScalePoint(const NewFirstDevice: TLongintPoint;
                                           const NewFirstActual: TDoublePoint);
begin
  FFirstScalePoint.MasterP := CurrentScaleSystem.DeviceToMaster(NewFirstDevice);
  FFirstScalePoint.ActualP := NewFirstActual;
  FFirstScalePoint.Defined := True;
  FPaintBox.Invalidate;
  //DoPaintScalePoints;
end;

procedure TPanZoomPanel.UnsetFirstScalePoint;
begin
  FFirstScalePoint.Defined := False;
  FPaintBox.Invalidate;
end;

procedure TPanZoomPanel.SetSkewPoint(const NewSkewPoint: TLongintPoint);
begin
  FSkewPoint.MasterP := CurrentScaleSystem.DeviceToMaster(NewSkewPoint);
  FSkewPoint.Defined := True;
  FPaintBox.Invalidate;
  //DoPaintScalePoints;
end;

procedure TPanZoomPanel.UnsetSkewPoint;
begin
  FSkewPoint.Defined := False;
  FPaintBox.Invalidate;
end;

procedure TPanZoomPanel.SetSecondScalePoint(const NewSecondDevice: TLongintPoint;
                                            const NewSecondActual: TDoublePoint);
begin
  FSecondScalePoint.MasterP := CurrentScaleSystem.DeviceToMaster(NewSecondDevice);
  FSecondScalePoint.ActualP := NewSecondActual;
  FSecondScalePoint.Defined := True;
  FPaintBox.Invalidate;
  //DoPaintScalePoints;
end;

procedure TPanZoomPanel.UnsetSecondScalePoint;
begin
  FSecondScalePoint.Defined := False;
  FPaintBox.Invalidate;
end;

procedure TPanZoomPanel.PaintScalePoints;
const
  ScalePointColor = clLime;

  procedure PaintScalePoint(const SP: TDefinableScalePoint);
  const
    ScalePointRadius = 4;
  var
    P: TLongintPoint;
  begin
    P := CurrentScaleSystem.MasterToDevice(SP.MasterP);
    FPaintBox.Canvas.Ellipse(P.X - ScalePointRadius, P.Y - ScalePointRadius,
                             P.X + ScalePointRadius, P.Y + ScalePointRadius);
  end;

begin
  FPaintBox.Canvas.Brush.Color := ScalePointColor;
  FPaintBox.Canvas.Pen.Color := ScalePointColor;
  if FFirstScalePoint.Defined then PaintScalePoint(FFirstScalePoint);
  if FSkewPoint.Defined then PaintScalePoint(FSkewPoint);
  if FSecondScalePoint.Defined then PaintScalePoint(FSecondScalePoint);
end;

procedure TPanZoomPanel.DoPaintScalePoints;
begin
  if FShowScalePoints then PaintScalePoints;
end;

procedure TPanZoomPanel.SetUseSpanningCursor(const Value: Boolean);
begin
  if FUseSpanningCursor <> Value then begin
    FUseSpanningCursor := Value;
    if FUseSpanningCursor then begin
      FSpanningCursor := TSpanningCursor.Create(FPaintBox);
      Repaint; // Keeps the spanning cursor from partially redrawing in ugly ways.
    end
    else begin
      FSpanningCursor.Free;
      FSpanningCursor := nil;
    end;
  end;
end;

procedure TPanZoomPanel.SetShowScalePoints(const Value: Boolean);
begin
  if FShowScalePoints <> Value then begin
    FShowScalePoints := Value;
    FPaintBox.Invalidate;
  end
end;

procedure TPanZoomPanel.ClearBitmap;
begin
  FBitmap.Free;
  FBitmap := nil;
  FPaintBox.Hide;
  FPaintBox.Update;
  UndefineScale;
  DoResize;
  //FPaintBox.ControlStyle := FPaintBox.ControlStyle - [csOpaque];
  FPaintBox.Show;
  Invalidate;
end;

procedure TPanZoomPanel.SetBitmapRef(const NewBitmap: TBitmap);
begin
  ClearBitmap;
  FBitmap := NewBitmap;
  //FPaintBox.ControlStyle := FPaintBox.ControlStyle + [csOpaque];
  ConvertIfFastPainting;
  OneToOne;
end;

procedure TPanZoomPanel.SetBitmap(const NewBitmap: TBitmap);
begin
  ClearBitmap;
  FBitmap := TBitmap.Create;
  FBitmap.Assign(NewBitmap);
  //FPaintBox.ControlStyle := FPaintBox.ControlStyle + [csOpaque];
  ConvertIfFastPainting;
  OneToOne;
end;

procedure TPanZoomPanel.Paint;
begin
  inherited;
end;

procedure TPanZoomPanel.Resize;
begin
  inherited;
  DoResize;
  if Assigned(FOnResize) then FOnResize(Self);
end;

function TPanZoomPanel.GetPaintBoxCursor: TCursor;
begin
  Result := FPaintBox.Cursor;
end;

procedure TPanZoomPanel.RedrawCursor;
var
  CursorPos: TPoint;
begin
  { This forces a redraw of the cursor and its new size. }
  if FUseSpanningCursor then begin
    GetCursorPos(CursorPos);
    CursorPos := FPaintBox.ScreenToClient(CursorPos);
    FSpanningCursor.Move(CursorPos);
  end;
end;

procedure TPanZoomPanel.SetPaintBoxCursor(const Value: TCursor);
var
  CursorPos: TPoint;
begin
  FPaintBox.Cursor := Value;
  GetCursorPos(CursorPos);
  SetCursorPos(CursorPos.X, CursorPos.Y);
end;

function TPanZoomPanel.GetPaintBoxCanvas: TCanvas;
begin
  Result := FPaintBox.Canvas;
end;

procedure TPanZoomPanel.ConvertIfFastPainting;
begin
  {
  if Assigned(FBitmap) and FUseFastPainting then begin
    (*$IFDEF PreDelphi3*)
      StupidHandle := FBitmap.Canvas.Handle;
    (*$ELSE*)
      FBitmap.PixelFormat := pfDevice;
    (*$ENDIF*)
  end;
  }
end;

procedure TPanZoomPanel.SetUseFastPainting(const Value: Boolean);
begin
  if FUseFastPainting <> Value then begin
    FUseFastPainting := Value;
    ConvertIfFastPainting;
  end;
end;

procedure TPanZoomPanel.ClearScalePoints;
begin
  FFirstScalePoint.Defined := False;
  FSkewPoint.Defined := False;
  FSecondScalePoint.Defined := False;
end;

procedure TPanZoomPanel.ScrollHorz(const Delta: Longint);
begin
  ScrollHorzOrVert(Delta, FHorzScrollBar);
end;

procedure TPanZoomPanel.ScrollVert(const Delta: Longint);
begin
  ScrollHorzOrVert(Delta, FVertScrollBar);
end;

procedure TPanZoomPanel.ZoomOut(const Factor: Double);
begin
  CurrentScaleSystem.WindowExtentP.X := Trunc(CurrentScaleSystem.WindowExtentP.X * Factor);
  CurrentScaleSystem.WindowExtentP.Y := Trunc(CurrentScaleSystem.WindowExtentP.Y * Factor);
  CheckZoomFactor;
  //if FIsZoomingIn or FIsPainting then Exit;
  ReconfigureCurrentSystem;
  DoResize;
  {if IsPainting then} CheckPenWidth;
  if FUseSpanningCursor then FPaintBox.Invalidate;
  if Assigned(FOnZoomOut) then FOnZoomOut(Self);
end;

procedure TPanZoomPanel.InternalSetIsZoomingIn(const Value: Boolean);
begin
  FIsZoomingIn := Value;
  if FIsZoomingIn then begin
    Movable := False;
    InternalSetIsPainting(False);
    InternalSetIsPanning(False);
    InternalSetIsSelecting(False);
    UseSpanningCursor := False;
    SetPaintBoxCursor(FZoomCursor);
    if FUseProportionalRubberBand then
      FRubberBand.SetProportional(UsableWidth, UsableHeight)
    else begin
      FRubberBand.SetNonProportional;
    end;
    if Assigned(FOnZoomInBegin) then FOnZoomInBegin(Self);
  end
  else begin
    FRubberBand.Cancel;
    if Assigned(FOnZoomInEnd) then FOnZoomInEnd(Self);
  end;
end;

procedure TPanZoomPanel.InternalSetIsPainting(const Value: Boolean);
begin
  FPenDown := False;
  if Value then begin
    FIsPainting := Value;
    Movable := False;
    FirstPoint := True;
    InternalSetIsPanning(False);
    InternalSetIsZoomingIn(False);
    InternalSetIsSelecting(False);
    UseSpanningCursor := False;
    CheckPenWidth;
    if Assigned(FOnPaintBegin) then FOnPaintBegin(Self);
  end
  else begin
    FIsPainting := Value;
    if Assigned(FOnPaintEnd) then FOnPaintEnd(Self);
  end;
end;

procedure TPanZoomPanel.InternalSetIsPanning(const Value: Boolean);
begin
  if Value then begin
    FIsPanning := Value;
    Movable := False;
    InternalSetIsPainting(False);
    InternalSetIsZoomingIn(False);
    InternalSetIsSelecting(False);
    UseSpanningCursor := False;
    SetPaintBoxCursor(crSizeAll);
    if Assigned(FOnPanBegin) then FOnPanBegin(Self);
  end
  else begin
    FIsPanning := Value;
    if Assigned(FOnPanEnd) then FOnPanEnd(Self);
  end;
end;

procedure TPanZoomPanel.SetIsZoomingIn(const Value: Boolean);
begin
  if FIsZoomingIn = Value then Exit;
  if Value then begin
    StoreState;
    InternalSetIsZoomingIn(Value);
  end
  else begin
    InternalSetIsZoomingIn(Value);
    RestoreState(siZooming);
  end;
end;

procedure TPanZoomPanel.SetIsPainting(const Value: Boolean);
begin
  if FIsPainting = Value then Exit;
  if Value then begin
    StoreState;
    InternalSetIsPainting(Value);
  end
  else begin
    InternalSetIsPainting(Value);
    RestoreState(siPainting);
  end;
end;

procedure TPanZoomPanel.SetIsPanning(const Value: Boolean);
begin
  if FIsPanning = Value then Exit;
  if Value then begin
    StoreState;
    InternalSetIsPanning(Value);
  end
  else begin
    InternalSetIsPanning(Value);
    RestoreState(siPanning);
  end;
end;

procedure TPanZoomPanel.SetUseProportionalRubberBand(const Value: Boolean);
begin
  if not FIsZoomingIn then FUseProportionalRubberBand := Value;
end;

procedure TPanZoomPanel.SetPenDown(const Value: Boolean);
begin
  FPenDown := Value;
  FirstPoint := FPenDown;
  if Assigned(FOnPenChanged) then FOnPenChanged(Self);
  BroadcastPenDownChange;
end;

procedure TPanZoomPanel.ClearPrevCoords;
begin
  PrevX := -1;
  PrevY := -1;
  FirstPoint := FIsPainting;
end;

procedure TPanZoomPanel.ReconfigureCurrentSystem;
begin
  if not ScaleDefined then Exit;
  CurrentScaleSystem.Define(OriginalScaleSystem.MasterP1,
                            OriginalScaleSystem.SkewP,
                            OriginalScaleSystem.MasterP2Prime,
                            OriginalScaleSystem.ActualP1,
                            OriginalScaleSystem.ActualP2,
                            OriginalScaleSystem.IsLogarithmicP);
  SetPaintBoxTransformation;
end;

procedure TPanZoomPanel.LoadFromStream(Stream: TStream);
var
  SD: Boolean; // ScaleDefined
  L: Longint;
  OSS, CSS: TScaleSystem;
begin
  Stream.Read(SD, SizeOf(SD));
  if SD then begin
    OSS := TScaleSystem.Create;
    try
      OSS.LoadFromStream(Stream);
      CSS := TScaleSystem.Create;
      try
        OSS.AssignTo(CSS);
        CSS.LoadFromStream(Stream);
        SetScaleSystems(OSS, CSS);
      finally
        CSS.Free;
      end;
    finally
      OSS.Free;
    end;
  end;
  DoResize;
  Stream.Read(L, SizeOf(L));
  ScrollHorz(L - FHorzScrollBar.Position);
  Stream.Read(L, SizeOf(L));
  ScrollVert(L - FVertScrollBar.Position);
end;

procedure TPanZoomPanel.SaveToStream(Stream: TStream);
var
  L: Longint;
begin
  Stream.Write(ScaleDefined, SizeOf(ScaleDefined));
  if ScaleDefined then begin
    OriginalScaleSystem.SaveToStream(Stream);
    CurrentScaleSystem.SaveToStream(Stream);
  end;
  L := FHorzScrollBar.Position;
  Stream.Write(L, SizeOf(L));
  L := FVertScrollBar.Position;
  Stream.Write(L, SizeOf(L));
end;

procedure TPanZoomPanel.DefineScale(const XIsLogarithmic, YIsLogarithmic: Boolean);
begin
  if not FFirstScalePoint.Defined or not FSkewPoint.Defined or
     not FSecondScalePoint.Defined then
  begin
    raise EDefineScale.Create('Scale points not completely defined.');
  end;
  try
    OriginalScaleSystem.WindowOriginP := CurrentScaleSystem.WindowOriginP;
    OriginalScaleSystem.WindowExtentP := CurrentScaleSystem.WindowExtentP;
    OriginalScaleSystem.ViewportOriginP := CurrentScaleSystem.ViewportOriginP;
    OriginalScaleSystem.ViewportExtentP := CurrentScaleSystem.ViewportExtentP;
    OriginalScaleSystem.Define(FFirstScalePoint.MasterP,
                               FSkewPoint.MasterP,
                               FSecondScalePoint.MasterP,
                               FFirstScalePoint.ActualP,
                               FSecondScalePoint.ActualP,
                               BPCreate(XIsLogarithmic, YIsLogarithmic));
    CurrentScaleSystem.Assign(OriginalScaleSystem);
    FScaleDefined := True;
    if Assigned(FOnScaleDefined) and not (csDestroying in ComponentState) then
    begin
      FOnScaleDefined(Self);
    end;
  except
    on E: EScaleSystem do begin
      FScaleDefined := False;
      raise EDefineScale.Create(E.Message);
    end;
  end;
end;

procedure TPanZoomPanel.SetScaleSystems(const Original, Current: TScaleSystem);
begin
  Original.AssignTo(FOriginalScaleSystem);
  Current.AssignTo(FCurrentScaleSystem);
  FScaleDefined := True;
  if Assigned(FOnScaleDefined) and not (csDestroying in ComponentState) then
  begin
    FOnScaleDefined(Self);
  end;
end;

procedure TPanZoomPanel.DefineScaleNoSkew(const XIsLogarithmic, YIsLogarithmic: Boolean);
begin
  SetSkewPoint(LPCreate(FSecondScalePoint.MasterP.X -
                        FFirstScalePoint.MasterP.X,
                        FFirstScalePoint.MasterP.Y));
  DefineScale(XIsLogarithmic, YIsLogarithmic);
end;

procedure TPanZoomPanel.UndefineScale;
begin
  if FScaleDefined then begin
    ClearScalePoints;
    FScaleDefined := False;
    if Assigned(FOnScaleUndefined) and not (csDestroying in ComponentState) then
      FOnScaleUndefined(Self);
  end;
end;

procedure TPanZoomPanel.OneToOne;
begin
  {CurrentScaleSystem.OneToOne;}
  with CurrentScaleSystem do begin
    WindowOriginP.X := 0;
    WindowOriginP.Y := 0;
    WindowExtentP.X := FBitmap.Width;
    WindowExtentP.Y := FBitmap.Height;
    ViewportOriginP := WindowOriginP;
    ViewportExtentP := WindowExtentP;
  end;
  ReconfigureCurrentSystem;
  DoResize;
  {if IsPainting then} CheckPenWidth;
end;

procedure TPanZoomPanel.SizeToImage;
var
  Borders: Integer;
  MaxWidth, MaxHeight: Longint;
begin
  with CurrentScaleSystem do begin
    MaxWidth := MulDiv(FBitmap.Width, ViewportExtentP.X, WindowExtentP.X);
    MaxHeight := MulDiv(FBitmap.Height, ViewportExtentP.Y, WindowExtentP.Y);
  end;
  Borders := TotalBorderWidth shl 1;
  SetBounds(Left, Top, MaxWidth + Borders, MaxHeight + Borders);
end;

function TPanZoomPanel.SetPaintBoxTransformation;
var
  TransformOK: Boolean;
begin
  with FPaintBox, CurrentScaleSystem do begin
    SetMapMode(Canvas.Handle, MM_ANISOTROPIC);
    TransformOK  := SetViewportOrgEx(Canvas.Handle, ViewportOriginP.X,
                                     ViewportOriginP.Y, nil) and
                    SetViewportExtEx(Canvas.Handle, ViewportExtentP.X,
                                     ViewportExtentP.Y, nil) and
                    SetWindowOrgEx(Canvas.Handle, WindowOriginP.X,
                                   WindowOriginP.Y, nil) and
                    SetWindowExtEx(Canvas.Handle, WindowExtentP.X,
                                   WindowExtentP.Y, nil);
  end;
  if not TransformOK then OneToOne;
end;

procedure TPanZoomPanel.UnsetPaintBoxTransformation;
begin
  with FPaintBox.Canvas do begin
    SetViewportOrgEx(Handle, 0, 0, nil);
    SetViewportExtEx(Handle, 1, 1, nil);
    SetWindowOrgEx(Handle, 0, 0, nil);
    SetWindowExtEx(Handle, 1, 1, nil);
  end;
end;

procedure TPanZoomPanel.PaintBitmap;
begin
  SetPaintBoxTransformation;
  with FPaintBox, CurrentScaleSystem do begin
    Canvas.Draw(MulDiv(Left, WindowExtentP.X, ViewportExtentP.X),
                MulDiv(Top, WindowExtentP.Y, ViewportExtentP.Y), FBitmap);
  end;
  UnsetPaintBoxTransformation;
end; { PaintBitmap }

procedure TPanZoomPanel.PaintBoxPaint(Sender: TObject);
begin
  if Assigned(FBitmap) then PaintBitmap else FPaintBox.Canvas.FillRect(FPaintBox.ClientRect);
  DoPaintScalePoints;
  //RedrawCursor;
  if Assigned(FOnPaintBoxPaint) then FOnPaintBoxPaint(Self);
  BroadcastPaintBoxPaint;
end;

procedure TPanZoomPanel.PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
                                          Shift: TShiftState; X, Y: Integer);
var
  ActualP: TDoublePoint;
begin
  if FIsZoomingIn or FIsSelecting and not FRubberBand.HasStarted then FRubberBand.Start(X, Y);
  if Assigned(FOnPaintBoxMouseDown) then begin
    ActualP := DeviceToActual(LPCreate(X, Y));
    FOnPaintBoxMouseDown(Self, Button, Shift, X, Y, ActualP.X, ActualP.Y);
  end;
end;

procedure TPanZoomPanel.PaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
                                        Shift: TShiftState; X, Y: Integer);
var
  ActualP: TDoublePoint;
begin
  if FIsZoomingIn or FIsSelecting and FRubberBand.HasStarted then FRubberBand.Stop(X, Y);
  if Assigned(FOnPaintBoxMouseUp) then begin
    ActualP := DeviceToActual(LPCreate(X, Y));
    FOnPaintBoxMouseUp(Self, Button, Shift, X, Y, ActualP.X, ActualP.Y);
  end;
end;

procedure TPanZoomPanel.PaintBoxMouseEnter(Sender: TObject);
begin
  //if FUseSpanningCursor then FSpanningCursor.Clear;
  if Assigned(FOnPaintBoxMouseEnter) then FOnPaintBoxMouseEnter(Self);
end;

procedure TPanZoomPanel.PaintBoxMouseLeave(Sender: TObject);
begin
  if FUseSpanningCursor then FSpanningCursor.Clear; 
  ClearPrevCoords;
  if Assigned(FOnPaintBoxMouseLeave) then FOnPaintBoxMouseLeave(Self);
end;

procedure TPanZoomPanel.PaintBoxClick(Sender: TObject);
begin
  if Assigned(FOnPaintBoxClick) then FOnPaintBoxClick(Sender);
end;

function TPanZoomPanel.DeviceToActual(const DeviceP: TLongintPoint): TDoublePoint;
begin
  if ScaleDefined then
    Result := CurrentScaleSystem.DeviceToActual(DeviceP)
  else begin
    Result := UndefinedDoublePoint;
  end;
end;

function TPanZoomPanel.ActualToDevice(const ActualP: TDoublePoint): TLongintPoint;
begin
  if ScaleDefined then
    Result := CurrentScaleSystem.ActualToDevice(ActualP)
  else begin
    Result := UndefinedLongintPoint;
  end;
end;

procedure TPanZoomPanel.ScrollHorzOrVert(const Delta: Longint; ScrollBar: TPageSizeScrollBar);
var
  DummyPos: Integer;
begin
  if ScrollBar.Visible then begin
    DummyPos := ScrollBar.Position + Delta;
    if DummyPos < ScrollBar.Min then
      DummyPos := ScrollBar.Min
    else begin
      DummyPos := Math.Min(ScrollBar.Max - ScrollBar.PageSize + 1, DummyPos);
    end;
    ScrollBar.Position := DummyPos;
    if Assigned(ScrollBar.OnScroll) then begin
      DummyPos := ScrollBar.Position;
      ScrollBar.OnScroll(ScrollBar, scTrack, DummyPos);
    end;
  end;
end;

procedure TPanZoomPanel.PaintBoxMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  Overshoot: Integer;
  DeviceP, MasterP: TLongintPoint;
  L, R, T, B: Integer;
  Rgn: HRGN;
begin
  DeviceP := LPCreate(X, Y);
  MasterP := CurrentScaleSystem.DeviceToMaster(DeviceP);
  if FIsZoomingIn or FIsSelecting then FRubberBand.Move(X, Y);
  if FIsPainting and FPenDown then begin
    with FPaintBox, CurrentScaleSystem do begin
      if FirstPoint then begin
        Windows.MoveToEx(FBitmap.Canvas.Handle, MasterP.X, MasterP.Y, nil);
        FirstPoint := False;
      end
      else begin
        if X < PrevX then begin
          L := X;
          R := PrevX;
        end
        else begin
          L := PrevX;
          R := X;
        end;
        if Y < PrevY then begin
          T := Y;
          B := PrevY;
        end
        else begin
          T := PrevY;
          B := Y;
        end;
        Windows.LineTo(FBitmap.Canvas.Handle, MasterP.X, MasterP.Y);
        Overshoot := 1 + Math.Ceil(FBitmap.Canvas.Pen.Width *
                                   (CurrentScaleSystem.ViewportExtentP.X /
                                    CurrentScaleSystem.WindowExtentP.X));
        Rgn := CreateRectRgn(L - Overshoot, T - Overshoot,
                             R + Overshoot, B + Overshoot);
        {L := }SelectClipRgn(Canvas.Handle, Rgn);
        PaintBitmap;
        SelectClipRgn(Canvas.Handle, 0);
        DeleteObject(Rgn);
      end;
    end;
  end;
  if FUseSpanningCursor then
    FSpanningCursor.Move(Point(X, Y));
  if IsPanning then begin
    if (PrevX >= 0) and (PrevY >= 0) then begin
      ScrollHorz(X - PrevX);
      ScrollVert(Y - PrevY);
    end;
  end;
  PrevX := X;
  PrevY := Y;
  if Assigned(FOnPaintBoxMouseMove) then begin
    FOnPaintBoxMouseMove(Self, Shift, DeviceP, MasterP);
  end;
end; { FPaintBoxMouseMove }

{
function TPanZoomPanel.MaxDeviceP: TLongintPoint;
begin
  with CurrentScaleSystem do begin
    Result.X := MulDiv(FPaintBox.Width, ViewportExtentP.X, WindowExtentP.X);
    Result.Y := MulDiv(FPaintBox.Height, ViewportExtentP.Y, WindowExtentP.Y);
  end;
end;
}

procedure TPanZoomPanel.SetupScrollBars;
var
  MaxWidth, MaxHeight: Longint;
  WasVisibleHorz, WasVisibleVert: Boolean;
  SPos, SPageSize, MinPageSize: Integer;

  procedure CheckHorzScrollBar;
  var
    OldHeight: Integer;
  begin
    with FHorzScrollBar do begin
      Visible := FPaintBox.Width < MaxWidth;
      if Visible then begin
        {
         The Align property changed from D2 to D3. Starting in D3, a control's
         dimensions would be actively changed to fit better. So, if the bevel
         of the panel is 3, the corresponding dimension of the control (for
         the vertical scroll bar, the width; for the horizontal, the height)
         would be shrunk that much. To get around this, we have to store the
         original dimension.
        }
        OldHeight := Height;
        Align := alBottom;
        Height := OldHeight;
      end;
    end;
  end;

  procedure CheckVertScrollBar;
  var
    OldWidth: Integer;
  begin
    with FVertScrollBar do begin
      Visible := FPaintBox.Height < MaxHeight;
      if Visible then begin
        {
         The Align property changed from D2 to D3. Starting in D3, a control's
         dimensions would be actively changed to fit better. So, if the bevel
         of the panel is 3, the corresponding dimension of the control (for
         the vertical scroll bar, the width; for the horizontal, the height)
         would be shrunk that much. To get around this, we have to store the
         original dimension.
        }
        OldWidth := Width;
        Align := alRight;
        Width := OldWidth;
      end;
    end;
  end;

begin
  with CurrentScaleSystem do begin
    MaxWidth := MulDiv(FBitmap.Width, ViewportExtentP.X, WindowExtentP.X);
    MaxHeight := MulDiv(FBitmap.Height, ViewportExtentP.Y, WindowExtentP.Y);
  end;
  WasVisibleHorz := FHorzScrollBar.Visible;
  WasVisibleVert := FVertScrollBar.Visible;
  FPaintBox.Align := alClient;
  CheckHorzScrollBar;
  CheckVertScrollBar;
  {
   If the horizontal scroll bar didn't need to appear but the vertical one
   did need to appear, the visibility of the vertical one might cause the
   horizontal one to need to appear (because FPaintBox.Width changed). So
   check again.
  }
  if not FHorzScrollBar.Visible then CheckHorzScrollBar;
  with FPaintBox do begin
    Align := alNone;
    if not FHorzScrollBar.Visible then Width := MaxWidth;
    if not FVertScrollBar.Visible then Height := MaxHeight;
  end;
  with FHorzScrollBar do begin
    if Visible then begin
      Align := alNone;
      if FVertScrollBar.Visible then Width := FVertScrollBar.Left - Left;
      with CurrentScaleSystem do begin
        SPageSize := 1 + MulDiv(FPaintBox.Width, WindowExtentP.X,
                                ViewportExtentP.X);
        MinPageSize := GetSystemMetrics(SM_CXHTHUMB);
        if SPageSize < MinPageSize then SPageSize := MinPageSize;
      end;
      if WasVisibleHorz then begin
        SPos := Position;
        if SPos > Succ(Max - SPageSize) then SPos := Succ(Max - SPageSize);
      end
      else begin
        SPos := 0;
      end;
      { NOTE: Delphi Help has the Min and Max parameters reversed. }
      PageSize := 2;
      LargeChange := 2;
      SetParams(SPos, Min, FBitmap.Width);
      PageSize := SPageSize;
      LargeChange := SPageSize;
      OnScroll(Self, scPosition, {scTrack,} SPos);
    end
    else begin
      CurrentScaleSystem.WindowOriginP.X := 0;
    end;
  end;
  with FVertScrollBar do begin
    if Visible then begin
      Align := alNone;
      if FHorzScrollBar.Visible then Height := FHorzScrollBar.Top - Top;
      with CurrentScaleSystem do begin
        SPageSize := 1 + MulDiv(FPaintBox.Height, WindowExtentP.Y,
                                ViewportExtentP.Y);
        MinPageSize := GetSystemMetrics(SM_CYVTHUMB);
        if SPageSize < MinPageSize then SPageSize := MinPageSize;
      end;
      if WasVisibleVert then begin
        SPos := Position;
        if SPos > Succ(Max - SPageSize) then SPos := Succ(Max - SPageSize);
      end
      else begin
        SPos := 0;
      end;
      { NOTE: Delphi Help has the Min and Max parameters reversed. }
      PageSize := 2;
      LargeChange := 2;
      SetParams(SPos, Min, FBitmap.Height);
      PageSize := SPageSize;
      LargeChange := SPageSize;
      OnScroll(Self, scPosition {scTrack}, SPos);
    end
    else begin
      CurrentScaleSystem.WindowOriginP.Y := 0;
    end;
  end;
  ReconfigureCurrentSystem;
end;

procedure TPanZoomPanel.DoResize;
begin
  if Assigned(FBitmap) then
    SetupScrollBars
  else begin
    FHorzScrollBar.Visible := False;
    FVertScrollBar.Visible := False;
  end;
  FPaintBox.Invalidate;
end;

const
  InvalidScrollCodes = [scEndScroll, scPosition];

procedure TPanZoomPanel.HorzScrollBarScroll(Sender: TObject;
                                            ScrollCode: TScrollCode;
                                            var ScrollPos: Integer);
begin
  if ScrollPos < FHorzScrollBar.Min then
    ScrollPos := FHorzScrollBar.Min
  else begin
    ScrollPos := Math.Min(FHorzScrollBar.Max - FHorzScrollBar.PageSize + 1, ScrollPos);
  end;
  CurrentScaleSystem.WindowOriginP.X := ScrollPos;
  if not (ScrollCode in InvalidScrollCodes) then FPaintBox.Invalidate;
end;

procedure TPanZoomPanel.VertScrollBarScroll(Sender: TObject;
                                            ScrollCode: TScrollCode;
                                            var ScrollPos: Integer);
begin
  if ScrollPos < FVertScrollBar.Min then
    ScrollPos := FVertScrollBar.Min
  else begin
    ScrollPos := Math.Min(FVertScrollBar.Max - FVertScrollBar.PageSize + 1, ScrollPos);
  end;
  CurrentScaleSystem.WindowOriginP.Y := ScrollPos;
  if not (ScrollCode in InvalidScrollCodes) then FPaintBox.Invalidate;
end;

procedure TPanZoomPanel.CheckZoomFactor;
const
  MaxFactor = 16;
begin
  if CurrentScaleSystem.ViewportExtentP.X / CurrentScaleSystem.WindowExtentP.X > MaxFactor then begin
    CurrentScaleSystem.WindowExtentP.X := CurrentScaleSystem.ViewportExtentP.X div MaxFactor;
  end;
  if CurrentScaleSystem.ViewportExtentP.Y / CurrentScaleSystem.WindowExtentP.Y > MaxFactor then begin
    CurrentScaleSystem.WindowExtentP.Y := CurrentScaleSystem.ViewportExtentP.Y div MaxFactor;
  end;
end;

procedure TPanZoomPanel.ZoomToArea(ARect: TRect);
var
  F1, F2: TLongintPoint;
  WasVisibleHorz, WasVisibleVert: Boolean;
begin
  with CurrentScaleSystem, ARect do begin
    F1 := DeviceToMaster(LPCreate(Left, Top));
    F2 := DeviceToMaster(LPCreate(Right, Bottom));
  end;

  WasVisibleHorz := FHorzScrollBar.Visible;
  WasVisibleVert := FVertScrollBar.Visible;

  with CurrentScaleSystem do begin
    WindowOriginP := F1;
    WindowExtentP.X := F2.X - F1.X;
    WindowExtentP.Y := F2.Y - F1.Y;
    ViewportExtentP.X := Self.UsableWidth;
    ViewportExtentP.Y := Self.UsableHeight;
    if FVertScrollBar.Visible then Dec(ViewportExtentP.X, FVertScrollBar.Width);
    if FHorzScrollBar.Visible then Dec(ViewportExtentP.Y, FHorzScrollBar.Height);
  end;

  CheckZoomFactor;

  DoResize;

  with CurrentScaleSystem do begin
    if FHorzScrollBar.Visible and not WasVisibleHorz then
      Dec(ViewportExtentP.Y, FHorzScrollBar.Height);
    if FVertScrollBar.Visible and not WasVisibleVert then
      Dec(ViewportExtentP.X, FVertScrollBar.Width);
  end;

  if (FHorzScrollBar.Visible and not WasVisibleHorz) or
     (FVertScrollBar.Visible and not WasVisibleVert) then
  begin
    DoResize;
  end;

  ScrollHorz(F1.X - FHorzScrollBar.Position);
  ScrollVert(F1.Y - FVertScrollBar.Position);
  FPaintBox.Invalidate;
  FRubberBand.Cancel;
  if FUseSpanningCursor then FPaintBox.Invalidate;
  if Assigned(OnZoomToArea) then OnZoomToArea(Self);
end;

procedure TPanZoomPanel.RubberBandStop(Sender: TObject; ARect: TRect);
var
  P: TLongintPoint;
  DeviceSelection: TRect;
  ActualSelection: TQuadrilateral;
begin
  with ARect do if (Right = Left) or (Bottom = Top) then begin
    FRubberBand.Cancel;
    Exit;
  end;
  {
   So the property setting routine knows which event to call, e.g.
   OnCancelZoomIn or OnEndZoomIn.
  }
  if FIsZoomingIn then begin
    ZoomToArea(ARect);
    //SetIsZoomingIn(False);
  end
  else if FIsSelecting then begin
    DeviceSelection := FRubberBand.CorrectBand;
    // Go around each point of DeviceSelection counterclockwise and create the
    // quadrilateral. The "actual" shape is not necessarily rectangular because
    // of skew angles.
    P.X := DeviceSelection.Left;
    P.Y := DeviceSelection.Top;
    // Top-left.
    ActualSelection[0] := FCurrentScaleSystem.DeviceToActual(P);
    P.Y := DeviceSelection.Bottom;
    // Bottom-left.
    ActualSelection[1] := FCurrentScaleSystem.DeviceToActual(P);
    P.X := DeviceSelection.Right;
    // Bottom-right.
    ActualSelection[2] := FCurrentScaleSystem.DeviceToActual(P);
    P.Y := DeviceSelection.Top;
    // Top-right.
    ActualSelection[3] := FCurrentScaleSystem.DeviceToActual(P);
    if Assigned(FOnSelectArea) then FOnSelectArea(Self, DeviceSelection, ActualSelection);
    //SetIsSelecting(False);
  end;
end;

procedure TPanZoomPanel.CreateAndSetPaintingCursor;
const
  CursorDim = 32;
  PaintingCursorHandle: HCursor = 0;
  (*$IFDEF Ver80*)
    CLR_INVALID = $FFFFFFFF;
  (*$ENDIF*)
var
  X, Y, Z: Integer;
  BitColor: TColorRef;
  ByteBMP, ByteMSK: Byte;
  Bitmap, Mask: array [0..127] of Byte;
  CursorBitmap: TBitmap;
  PWidth: Longint;
  ActualGray, ActualYellow: TColorRef; {TColor;}
begin
  if not RunningNT then begin
    if (GetSystemMetrics(SM_CXCURSOR) <> CursorDim) or
       (GetSystemMetrics(SM_CYCURSOR) <> CursorDim) then
    begin
      Exit;
    end;
  end;
  with CurrentScaleSystem do begin
    PWidth := 2 + MulDiv(FBitmap.Canvas.Pen.Width, ViewportExtentP.X,
                         WindowExtentP.X);
  end;
  Screen.Cursors[crPainting] := 0;
  if PaintingCursorHandle <> 0 then DestroyCursor(PaintingCursorHandle);
  CursorBitmap := TBitmap.Create;
  with CursorBitmap do begin
    Width := CursorDim;
    Height := Width;
    with Canvas do begin
      Pixels[0, 0] := clBlack;
      ActualGray := Pixels[0, 0];
      Pixels[0, 0] := clYellow;
      ActualYellow := Pixels[0, 0];
    end;
    { Make bitmap entirely gray, i.e. transparent. }
    SetPenAndBrush(Canvas, ActualGray, pmCopy, psSolid, 1, ActualGray, bsSolid);
    Canvas.Rectangle(0, 0, Width, Height);
    { Now draw a square to show the pen size. }
    SetPenAndBrush(Canvas, ActualYellow, pmCopy, psSolid, 1, ActualYellow,
                   bsClear);
    Canvas.Rectangle((Width - PWidth) div 2, (Height - PWidth) div 2,
                     (Width + PWidth) div 2, (Height + PWidth) div 2);
  end;

  for Y := 0 to 31 {CursorBitmap.Height - 1} do begin
    for X := 0 to 3 {CursorBitmap.Width div 8 - 1} do begin
      ByteBMP := 0;
      ByteMSK := 0;
      for Z := 0 to 7 do begin
        BitColor := GetPixel(CursorBitmap.Canvas.Handle, (X * 8) + Z, Y);
        if BitColor <> CLR_INVALID then begin
          ByteBMP := ByteBMP shl 1 or 1;
          ByteMSK := ByteMSK shl 1 or Byte(BitColor = ActualYellow);
        end;
      end;
      Bitmap[Y * 4 + X] := ByteBMP;
      Mask[Y * 4 + X] := ByteMSK;
    end;
  end;

  PaintingCursorHandle := CreateCursor(HInstance, CursorBitmap.Width div 2,
                                       CursorBitmap.Height div 2,
                                       CursorBitmap.Width, CursorBitmap.Height,
                                       @Bitmap, @Mask);

  CursorBitmap.Free;

  Screen.Cursors[crPainting] := PaintingCursorHandle;

  SetPaintBoxCursor(crPainting);
end; { CreateAndSetPaintingCursor }

function TPanZoomPanel.GetPenWidth: Integer;
begin
  Result := FBitmap.Canvas.Pen.Width;
end;

procedure TPanZoomPanel.SetPenWidth(Value: Integer);
var
  Max: Integer;
begin
  if Assigned(FBitmap) then begin
    if ScaleDefined then
      Max := Trunc(31 * CurrentScaleSystem.WindowExtentP.X / CurrentScaleSystem.ViewportExtentP.X)
    else begin
      Max := 31;
    end;
    if Value < 1 then Value := 1 else if Value > Max then Value := Max;
    FBitmap.Canvas.Pen.Width := Value;
    if IsPainting then CreateAndSetPaintingCursor;
    BroadcastPenWidthChange;
  end;
end;

procedure TPanZoomPanel.CheckPenWidth;
begin
  SetPenWidth(FBitmap.Canvas.Pen.Width);
end;

function TPanZoomPanel.GetRubberBandWidth: Integer;
begin
  Result := FRubberBand.BandWidth;
end;

procedure TPanZoomPanel.SetRubberBandWidth(const NewRubberBandWidth: Integer);
begin
  FRubberBand.BandWidth := NewRubberBandWidth;
end;


{ -----------------------------------------------------------------------------}


constructor TPsRubberBand.Create(AOwner: TComponent);
begin
  inherited;
  FHasStarted := False;
  FBandWidth := DefBandWidth;
  Brush := TBrush.Create;
  with Brush do begin
    Color := clWhite;
    Style := bsSolid;
  end;
end;

destructor TPsRubberBand.Destroy;
begin
  Brush.Free;
  inherited;
end;

procedure TPsRubberBand.SetProportional(const APropWidth, APropHeight: Longint);
begin
  FProportional := True;
  FPropWidth := APropWidth;
  FPropHeight := APropHeight;
end;

procedure TPsRubberBand.SetNonProportional;
begin
  FProportional := False;
end;

procedure TPsRubberBand.Start(const X, Y: Integer);
begin
  if FHasStarted then Exit;
  Band := Rect(X, Y, X, Y);
  FHasStarted := True;
end;

procedure TPsRubberBand.Move(const X, Y: Integer);
begin
  if not FHasStarted then Exit;
  OldBrush := FCanvas.Brush;
  FCanvas.Brush := Brush;
  Paint;
  UpdateBand(X, Y);
  Paint;
  FCanvas.Brush := OldBrush;
end;

procedure TPsRubberBand.Cancel;
begin
  if not FHasStarted then Exit;
  OldBrush := FCanvas.Brush;
  FCanvas.Brush := Brush;
  Paint;
  FCanvas.Brush := OldBrush;
  FHasStarted := False;
  if Assigned(FOnCancel) then FOnCancel(Self);
end;

procedure TPsRubberBand.Stop(const X, Y: Integer);
begin
  if not FHasStarted then Exit;
  OldBrush := FCanvas.Brush;
  FCanvas.Brush := Brush;
  Paint;
  FCanvas.Brush := OldBrush;
  UpdateBand(X, Y);
  FHasStarted := False;
  if Assigned(FOnStop) then FOnStop(Self, CorrectBand);
end;

function TPsRubberBand.CorrectBand: TRect;
begin
  with Band do begin
    if Left > Right then begin
      Result.Left := Right;
      Result.Right := Left;
    end
    else begin
      Result.Left := Left;
      Result.Right := Right;
    end;
    if Top > Bottom then begin
      Result.Top := Bottom;
      Result.Bottom := Top;
    end
    else begin
      Result.Top := Top;
      Result.Bottom := Bottom;
    end;
  end;
end;

procedure TPsRubberBand.SetBandWidth(const NewBandWidth: Integer);
begin
  if FHasStarted or (NewBandWidth < 1) then Exit;
  FBandWidth := NewBandWidth;
end;

procedure TPsRubberBand.Paint;
var
  I: Integer;
  CB: TRect;
begin
  CB := CorrectBand;
  for I := 1 to FBandWidth do begin
    Windows.DrawFocusRect(FCanvas.Handle, CB);
    with CB do begin
      Dec(Left);
      Dec(Top);
      Inc(Right);
      Inc(Bottom);
    end;
  end;
end;

procedure TPsRubberBand.UpdateBand(const X, Y: Integer);
var
  W, H: Longint;
begin
  with Band do begin
    if FProportional then begin
      W := MulDiv(Abs(Y - Top), FPropWidth, FPropHeight);
      H := Abs(X - Left);
      if W < H then begin
        if X < Left then W := -W;
        BottomRight := Point(Left + W, Y);
      end
      else begin
        H := MulDiv(H, FPropHeight, FPropWidth);
        if Y < Top then H := -H;
        BottomRight := Point(X, Top + H);
      end;
    end
    else begin
      BottomRight := Point(X, Y);
    end;
  end;
end;


{ -----------------------------------------------------------------------------}


constructor TResizablePanel.Create(AOwner: TComponent);
begin
  inherited;
  FMovable := DefMovable;
  FSizable := DefSizable;
  FSizeHandleWidth := DefSizeHandleWidth;
end;

destructor TResizablePanel.Destroy;
begin
  inherited;
end;

procedure TResizablePanel.WMNCHitTest(var Mess: TWMNCHitTest);
const
  MovableStatus: array [Boolean] of Longint = (HTCLIENT, HTCAPTION);
var
  P: TPoint;
begin
  inherited;
  if csDesigning in ComponentState then begin
    Parent.Dispatch(Mess);
    Exit;
  end;
  with Mess do begin
    if FSizable and (Align = alNone) then begin
      P := ScreenToClient(Point(XPos, YPos));
      if (P.X < FSizeHandleWidth) and (P.Y < FSizeHandleWidth) then
        Result := HTTOPLEFT
      else if (P.X < FSizeHandleWidth) and (Height - P.Y < FSizeHandleWidth) then
        Result := HTBOTTOMLEFT
      else if P.X < FSizeHandleWidth then
        Result := HTLEFT
      else if (Width - P.X < FSizeHandleWidth) and (P.Y < FSizeHandleWidth) then
        Result := HTTOPRIGHT
      else if (Width - P.X < FSizeHandleWidth) and
              (Height - P.Y < FSizeHandleWidth) then
      begin
        Result := HTBOTTOMRIGHT;
      end
      else if Width - P.X < FSizeHandleWidth then
        Result := HTRIGHT
      else if P.Y < FSizeHandleWidth then
        Result := HTTOP
      else if Height - P.Y < FSizeHandleWidth then
        Result := HTBOTTOM;
    end;
    if Result = 0 then Result := MovableStatus[(Align = alNone) and FMovable];
  end;
end;

function TResizablePanel.GetTotalBorderWidth: Integer;
begin
  Result := Ord(BorderStyle = bsSingle) +
            Ord(BevelOuter <> bvNone) * BevelWidth +
            Ord(BevelInner <> bvNone) * BorderWidth +
            Ord(BevelInner <> bvNone) * BevelWidth;
end;

function TResizablePanel.GetUsableWidth: Integer;
begin
  Result := ClientWidth - TotalBorderWidth shl 1;
end;

function TResizablePanel.GetUsableHeight: Integer;
begin
  Result := ClientHeight - TotalBorderWidth shl 1;
end;

{ -----------------------------------------------------------------------------}


{ TPanZoomPanelState }

procedure TPanZoomPanelState.Read(Panel: TPanZoomPanel);
begin
  FMovable := Panel.Movable;
  FIsPainting := Panel.IsPainting;
  FIsPenDown := Panel.PenDown;
  FIsPanning := Panel.IsPanning;
  FIsZooming := Panel.IsZoomingIn;
  FIsSelecting := Panel.IsSelecting;
  FPaintBoxCursor := Panel.FPaintBox.Cursor;
  FUseSpanningCursor := Panel.UseSpanningCursor;
end;

procedure TPanZoomPanel.RestoreState(const IgnoreItem: TPanZoomPanelStateItem);
var
  S: TPanZoomPanelState;
begin
  if FStates.Count = 0 then Exit;
  S := FStates.Last;
  FStates.Count := FStates.Count - 1;
  if IgnoreItem <> siMovable then Movable := S.FMovable;
  if IgnoreItem <> siPainting then begin
    InternalSetIsPainting(S.FIsPainting);
    SetPenDown(S.FIsPenDown);
  end;
  if IgnoreItem <> siPanning then InternalSetIsPanning(S.FIsPanning);
  if IgnoreItem <> siZooming then InternalSetIsZoomingIn(S.FIsZooming);
  if IgnoreItem <> siSelecting then InternalSetIsSelecting(S.FIsSelecting);
  if IgnoreItem <> siPaintBoxCursor then SetPaintBoxCursor(S.FPaintBoxCursor);
  UseSpanningCursor := S.FUseSpanningCursor;
  S.Free;
end;

procedure TPanZoomPanel.StoreState;
var
  S: TPanZoomPanelState;
begin
  S := TPanZoomPanelState.Create;
  S.Read(Self);
  FStates.Add(S);
end;

procedure TPanZoomPanel.InternalSetIsSelecting(const Value: Boolean);
begin
  FIsSelecting := Value;
  if FIsSelecting then begin
    Movable := False;
    InternalSetIsZoomingIn(False);
    InternalSetIsPainting(False);
    InternalSetIsPanning(False);
    UseSpanningCursor := False;
    SetPaintBoxCursor(FZoomCursor);
    FRubberBand.SetNonProportional;
    if Assigned(FOnSelectBegin) then FOnSelectBegin(Self);
  end
  else begin
    FRubberBand.Cancel;
    if Assigned(FOnSelectEnd) then FOnSelectEnd(Self);
  end;
end;

procedure TPanZoomPanel.SetIsSelecting(const Value: Boolean);
begin
  if FIsSelecting = Value then Exit;
  if Value then begin
    StoreState;
    InternalSetIsSelecting(Value);
  end
  else begin
    InternalSetIsSelecting(Value);
    RestoreState(siSelecting);
  end;
end;

constructor TSpanningCursor.Create(PaintBox: TPaintBox);
begin
  inherited Create;
  FHasPosition := False;
  FPaintBox := PaintBox;
end;

destructor TSpanningCursor.Destroy;
begin
  if FHasPosition then Paint;
  inherited;
end;

procedure TSpanningCursor.Move(const Position: TPoint);
begin
  if FHasPosition and (FPosition.X = Position.X) and (FPosition.Y = Position.Y) then Exit;
  if FHasPosition then Paint;
  FPosition := Position;
  FHasPosition := True;
  Paint;
end;

procedure TSpanningCursor.MoveSilent(const Position: TPoint);
begin
  //if FHasPosition and (FPosition.X = Position.X) and (FPosition.Y = Position.Y) then Exit;
  FPosition := Position;
  FHasPosition := True;
  Paint;
end;

procedure TSpanningCursor.Paint;
var
  H: THandle;
begin
  FPaintBox.Canvas.Brush.Style := bsSolid;
  FPaintBox.Canvas.Brush.Color := clWhite;
  H := FPaintBox.Canvas.Handle;
  Windows.DrawFocusRect(H, Rect(0, FPosition.Y - 1, FPaintBox.Width, FPosition.Y + 1));
  Windows.DrawFocusRect(H, Rect(FPosition.X - 1, 0, FPosition.X + 1, FPaintBox.Height));
end;

procedure TSpanningCursor.Clear;
begin
  if FHasPosition then Paint;
  FHasPosition := False;
end;

function TPanZoomPanel.IsRubberBanding: Boolean;
begin
  Result := FRubberBand.HasStarted;
end;

procedure TPanZoomPanel.DoEnter;
begin
  inherited;
end;

procedure TPanZoomPanel.UpdatePaintingCursor;
begin
  CreateAndSetPaintingCursor;
end;

procedure TPanZoomPanel.Attach(const Observer: IPanZoomPanelObserver);
begin
  if not IsObserved(Observer) then FObservers.Add(Observer);
end;

procedure TPanZoomPanel.Detach(const Observer: IPanZoomPanelObserver);
begin
  FObservers.Remove(Observer);
end;

function TPanZoomPanel.IsObserved(const Observer: IPanZoomPanelObserver): Boolean;
begin
  Result := FObservers.IndexOf(Observer) > -1;
end;

procedure TPanZoomPanel.BroadcastPaintBoxPaint;
var
  I: Integer;
begin
  for I := 0 to FObservers.Count - 1 do begin
    (FObservers[I] as IPanZoomPanelObserver).OnPaintBoxPaint(Self);
  end;
end;

procedure TPanZoomPanel.BroadcastPenDownChange;
var
  I: Integer;
begin
  for I := 0 to FObservers.Count - 1 do begin
    (FObservers[I] as IPanZoomPanelObserver).OnPenDownChange(Self);
  end;
end;

procedure TPanZoomPanel.BroadcastPenWidthChange;
var
  I: Integer;
begin
  for I := 0 to FObservers.Count - 1 do begin
    (FObservers[I] as IPanZoomPanelObserver).OnPenWidthChange(Self);
  end;
end;

initialization
(*$IFDEF Win32*)
  OsVersionInfo.dwOSVersionInfoSize := SizeOf(OsVersionInfo);
(*$ENDIF*)
  RunningNT := (*$IFDEF Win32*)
                 GetVersionEx(OsVersionInfO) and
                 (OsVersionInfo.dwPlatformId = VER_PLATFORM_WIN32_NT);
               (*$ELSE*)
                 False;
               (*$ENDIF*)

end.
