unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  PanZoom, StdCtrls, ExtCtrls, ComCtrls, DoublePoints;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    cmdZoomIn: TButton;
    cmdOneToOne: TButton;
    cmdZoomOut: TButton;
    Splitter1: TSplitter;
    cmdSizeToImage: TButton;
    cmdPaint: TButton;
    lblXY: TLabel;
    lblStatus: TLabel;
    chkProportionalZoom: TCheckBox;
    chkAligned: TCheckBox;
    PanZoomPanel1: TPanZoomPanel;
    cmdLoadImage: TButton;
    OpenDialog1: TOpenDialog;
    chkResizable: TCheckBox;
    lblPenWidth: TLabel;
    txtPenWidth: TEdit;
    udPenWidth: TUpDown;
    procedure cmdZoomInClick(Sender: TObject);
    procedure cmdOneToOneClick(Sender: TObject);
    procedure cmdZoomOutClick(Sender: TObject);
    procedure cmdSizeToImageClick(Sender: TObject);
    procedure cmdPaintClick(Sender: TObject);
    procedure chkAlignedClick(Sender: TObject);
    procedure PanZoomPanel1ZoomInBegin(Sender: TObject);
    procedure PanZoomPanel1PaintBegin(Sender: TObject);
    procedure cmdLoadImageClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure chkResizableClick(Sender: TObject);
    procedure txtPenWidthChange(Sender: TObject);
    procedure PanZoomPanel1MouseMove(Sender: TObject; Shift: TShiftState;
      const DeviceP, MasterP: TLongintPoint);
    procedure PanZoomPanel1Click(Sender: TObject);
  private
    procedure CheckWidgets;
    procedure SetPenWidth;
  public
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.cmdZoomInClick(Sender: TObject);
begin
  PanZoomPanel1.UseProportionalRubberBand := chkProportionalZoom.Checked;
  PanZoomPanel1.IsZoomingIn := not PanZoomPanel1.IsZoomingIn;
end;

procedure TForm1.cmdOneToOneClick(Sender: TObject);
begin
  PanZoomPanel1.OneToOne;
end;

procedure TForm1.cmdZoomOutClick(Sender: TObject);
var
  Factor: string;
begin
  Factor := '1.0';
  if InputQuery('Zoom Out', 'Enter factor:', Factor) then begin
    PanZoomPanel1.ZoomOut(StrToFloat(Factor));
  end;
end;

procedure TForm1.cmdSizeToImageClick(Sender: TObject);
begin
  PanZoomPanel1.SizeToImage;
end;

procedure TForm1.cmdPaintClick(Sender: TObject);
begin
  PanZoomPanel1.IsPainting := not PanZoomPanel1.IsPainting;
end;

procedure TForm1.chkAlignedClick(Sender: TObject);
const
  Aligns: array [Boolean] of TAlign = (alNone, alClient);
begin
  if chkAligned.Checked then chkResizable.Checked := False;
  PanZoomPanel1.Align := Aligns[chkAligned.Checked];
end;

procedure TForm1.PanZoomPanel1ZoomInBegin(Sender: TObject);
const
  ZoomingStatus: array [Boolean] of string = ('', 'Zooming');
  ZoomInCaptions: array [Boolean] of string = ('Zoom &In', '&Cancel');
begin
  lblStatus.Caption := ZoomingStatus[PanZoomPanel1.IsZoomingIn];
  cmdZoomIn.Caption := ZoomInCaptions[PanZoomPanel1.IsZoomingIn];
  CheckWidgets;
end;

procedure TForm1.PanZoomPanel1PaintBegin(Sender: TObject);
const
  PaintingStatus: array [Boolean] of string = ('', 'Painting. Click on image to toggle pen');
  PaintCaptions: array [Boolean] of string = ('&Paint', '&Cancel');
begin
  lblStatus.Caption := PaintingStatus[PanZoomPanel1.IsPainting];
  cmdPaint.Caption := PaintCaptions[PanZoomPanel1.IsPainting];
  CheckWidgets;
end;

procedure TForm1.CheckWidgets;
begin
  cmdLoadImage.Enabled := not Assigned(PanZoomPanel1.Bitmap);
  cmdZoomIn.Enabled := Assigned(PanZoomPanel1.Bitmap) and
                       not PanZoomPanel1.IsPainting;
  cmdOneToOne.Enabled := Assigned(PanZoomPanel1.Bitmap) and
                         not PanZoomPanel1.IsZoomingIn and
                         not PanZoomPanel1.IsPainting;
  cmdZoomOut.Enabled := Assigned(PanZoomPanel1.Bitmap) and
                        not PanZoomPanel1.IsZoomingIn and
                        not PanZoomPanel1.IsPainting;
  cmdSizeToImage.Enabled := Assigned(PanZoomPanel1.Bitmap) and
                            not PanZoomPanel1.IsZoomingIn and
                            not PanZoomPanel1.IsPainting;
  cmdPaint.Enabled := Assigned(PanZoomPanel1.Bitmap) and
                      not PanZoomPanel1.IsZoomingIn;
end;

procedure TForm1.cmdLoadImageClick(Sender: TObject);
var
  Bitmap: TBitmap;
begin
  if OpenDialog1.Execute then begin
    Bitmap := TBitmap.Create;
    try
      Bitmap.LoadFromFile(OpenDialog1.FileName);
      PanZoomPanel1.Bitmap := Bitmap;
      PanZoomPanel1.SetFirstScalePoint(LPCreate(0, Bitmap.Height),
                                       DPCreate(0, 0));
      PanZoomPanel1.SetSecondScalePoint(LPCreate(Bitmap.Width, 0),
                                        DPCreate(100, 100));
      PanZoomPanel1.DefineScaleNoSkew(False, False);
    finally
      Bitmap.Free;
    end;
  end;
  CheckWidgets;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Action := caFree;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  CheckWidgets;
end;

procedure TForm1.chkResizableClick(Sender: TObject);
begin
  PanZoomPanel1.Sizable := chkResizable.Checked;
end;

procedure TForm1.SetPenWidth;
begin
  PanZoomPanel1.PenWidth := StrToIntDef(txtPenWidth.Text, 1);
end;

procedure TForm1.txtPenWidthChange(Sender: TObject);
begin
  SetPenWidth;
end;

procedure TForm1.PanZoomPanel1MouseMove(Sender: TObject;
  Shift: TShiftState; const DeviceP, MasterP: TLongintPoint);
const
  sFormat = '0.00';
var
  ActualP: TDoublePoint;
  S: string;
begin
  S := '(' + IntToStr(DeviceP.X) + ', ' + IntToStr(DeviceP.Y) + ')';
  if PanZoomPanel1.ScaleDefined then begin
    ActualP := PanZoomPanel1.DeviceToActual(DeviceP);
    S := S + ' : (' + FormatFloat(sFormat, ActualP.X) + ', ' +
                      FormatFloat(sFormat, ActualP.Y) + ')';
  end;
  lblXY.Caption := S;
end;


procedure TForm1.PanZoomPanel1Click(Sender: TObject);
begin
  if PanZoomPanel1.IsPainting then begin
    PanZoomPanel1.PenDown := not PanZoomPanel1.PenDown;
  end;
end;

end.
